import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import j3dui.utils.Debug;
import j3dui.utils.*;
import j3dui.utils.app.*;
import j3dui.utils.blocks.*;
import j3dui.utils.objects.*;
import j3dui.control.*;
import j3dui.control.mappers.*;
import j3dui.control.actuators.groups.*;
import j3dui.feedback.elements.*;
import j3dui.navigate.*;

/**
Example demonstrating orbit and fly camera navigation
techniques.
<P>
The left display is from a first person 'fly' camera.
To operate this camera effectively you have to ignore
the other display and imagine that you are flying an
airplane.  The right display is from a second person
'orbit' camera.  The fly camera is represented in it
by an airplane model.  The orbit camera LAP is
represented in it as a small blue and magenta thing.
Sometimes the last modifier key sticks -- press and
release it to unstick it.
<P>
This example was developed after the book was printed and, as
such, is not described in the book.

@author Jon Barrilleax,
copyright (c) 2000 Jon Barrilleaux,
All Rights Reserved.
*/

public class CameraNavigating extends DoubleFrame {
	
	public static void main(String[] argv) {
		Debug.loadAllProperties();
		new CameraNavigating();
	}
	
	// public interface =========================================
	
	public CameraNavigating() {
		setTitle("Example - CameraNavigating");
		
		new HelpDialog(this, "Help - CameraNavigating",
		 "Example demonstrating the orbit and fly camera " +
		 "navigation techniques." +
		 "\n" +
		 "\n" +
		 ">>> The left display is from a first person " +
		 "'fly' camera.  To operate this camera effectively " +
		 "you have to ignore the other display and imagine " +
		 "that you are flying an airplane." +
		 "\n" +
		 "> Left-button drag to roll and pitch view." +
		 "\n" +
		 "> Left-right arrow keys to yaw the view." +
		 "\n" +
		 "> Up-down arrow keys to accelerate the view." +
		 "\n" +
		 "> Escape key to stop the view." +
		 "\n" +
		 "\n" +
		 ">>> The right display is from a second person " +
		 "'orbit' camera.  The fly camera is represented in " +
		 "it by an airplane model.  The orbit camera LAP is " +
		 "represented in it as a small blue and magenta " +
		 "thing." +
		 "\n" +
		 "> Left-button drag or arrow keys to " +
		 "tilt and turn view." +
		 "\n" +
		 "> Left-button-ctrl drag or arrow keys-ctrl to " +
		 "change LFO-Z." +
		 "\n" +
		 "> Left-button-shift-ctrl drag or arrow " +
		 "keys-shift-ctrl to slide LAP in X-Z using DRM." +
		 "\n" +
		 "\n" +
		 ">>> Sometimes the last modifier key sticks -- " +
		 "press and release it to unstick it." +
		 "\n" +
		 "\n" +
		 "This example is not described in the book because " +
		 "it was developed after the book was printed." +
		 "\n" +
		 "copyright (c) 1999-2000 Jon Barrilleaux, " +
		 "All Rights Reserved."
		);

		// configure views for large area
		AppView flyView = getViewLeft();
		flyView.setViewClipDist(new Vector2d(.1, 200));
		
		AppView orbitView = getViewRight();
		orbitView.setViewClipDist(new Vector2d(.1, 200));
		
		// build fly camera and controls
		FlyCamera flyCamera = new FlyCamera(flyView);
		FlyCameraControl flyControl = new FlyCameraControl(
		 flyView.getDisplay(), flyCamera);
		
		/// make cancel work from both views
		flyControl.getThrustCanceler().addEventSource(
		 orbitView.getDisplay());

		/// attach camera to world		
		getWorld().addViewNode(flyCamera); 

		/// build airplane avatar at LFP
		flyView.addNode(ModelLoader.loadModelFromName(
		 "plane.wrl"));

		/// limit movement to volume over the earth		
		flyCamera.getThrustActuator().getPlugin().
		 setTargetClamp(Mapper.DIM_X|Mapper.DIM_Z,
		 new Vector2d(-50, 50));
		flyCamera.getThrustActuator().getPlugin().
		 setTargetClamp(Mapper.DIM_Y,
		 new Vector2d(2, 40));

		/// start at the south looking north		
		flyCamera.getThrustActuator().initActuation(
		 new Vector4d(0, 10, 40, 0));

		// build orbit camera and controls
		OrbitCamera orbitCamera = new OrbitCamera(orbitView);
		OrbitCameraControl orbitControl = new OrbitCameraControl(
		 orbitView.getDisplay(), orbitCamera); 
		
		/// attach camera to fly camera's unoriented LFP		
		flyCamera.getThrustGroup().addChild(orbitCamera); 

		/// build LAP thing
		AffineGroup lapThing = new AffineGroup(new TestThing());
		lapThing.getScale().initActuation(
		 new Vector4d(.25, .25, .25, 0));
		
		orbitCamera.getLapGroup().addChild(lapThing); 

		/// start looking from the south west
		orbitCamera.getOrbitActuator().initActuation(
		 new Vector4d(-Math.PI/6, -Math.PI/8, 0, 0));

		// build world ground
		/// build small textured earth rect
		TextureRect earth = new TextureRect(true,
		 new Color3f(1, 1, 1), 0, "earth.jpg",
		 Elements.SIDE_TOP, 100.0, 100.0);
		 
		/// build large untextured water rect
		TextureRect waterRect = new TextureRect(true,
		 new Color3f(.05f, .25f, .66f), 0, null,
		 Elements.SIDE_TOP, 500.0, 500.0);
		AffineGroup water = new AffineGroup(waterRect);
		
		/// make water non-coplanar with earth
		water.getTranslation().initActuation(
		 new Vector4d(0, 0, -1, 0));

		/// compose ground and flip it up		
		AffineGroup ground = new AffineGroup();
		ground.addNode(earth);	 
		ground.addNode(water);	 
		ground.getAxisAngle().initActuation(
		 new Vector4d(1, 0, 0, -Math.PI/2.0));
		
		getWorld().addSceneNode(ground);

		// build world thing at origin
		getWorld().addSceneNode(new TestThing(
		 new Color3f(1f, 1f, 0f), new Color3f(1f, 0f, 0f)));

		// build world light
		OrbitGroup sun = new OrbitGroup(new TestLight(
		 new DirectionalLight()));
		sun.initActuation(new Vector4d(
		 -Math.PI/6, -Math.PI/8, 0, 0));
		
		getWorld().addSceneNode(sun);

		// build world sky
		/// build large textured cone
		TextureCone sky = new TextureCone(
		 Elements.SIDE_BOTTOM, 200, 100);
		sky.setTexture("sky.jpg");
		
		getWorld().addSceneNode(sky);
		
		// done
		showFrame();
	}
			
	// personal body ============================================

}