import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.dnd.*;
import java.awt.datatransfer.*;
import java.io.IOException;
import javax.media.j3d.*;
import javax.vecmath.*;

import j3dui.utils.Debug;
import j3dui.utils.*;
import j3dui.utils.app.*;
import j3dui.control.inputs.*;
import j3dui.control.mappers.*;

/**
A 3D view display capable of initiating and accepting drag and
drop operations.

@author Jon Barrilleaux,
copyright (c) 1999 Jon Barrilleaux,
All Rights Reserved.
*/
public class DndAppDisplay extends AppDisplay
 implements DragSourceListener, DragGestureListener,
 DropTargetListener {
	
	// public interface =========================================
	
	/**
	Construct a DndAppDisplay with default drag and drop actions.
	@param world Host world serving as the root for all scene
	picking and object instantiation.
	@param dragSourceList List containing references to nodes
	that can act as drag source objects.
	@param dropTargetList List containing references to nodes
	that can act as drop target objects.
	@param dropSourceList List containing references to nodes
	that can act as drop source objects.
	*/
	public DndAppDisplay(AppWorld world,
	 ArrayList dragSourceList, ArrayList dropTargetList,
	 ArrayList dropSourceList) {
	 
		this(world, dragSourceList, dropTargetList,
		 dropSourceList,
		 DnDConstants.ACTION_COPY_OR_MOVE |
		 DnDConstants.ACTION_LINK,
		 DnDConstants.ACTION_COPY_OR_MOVE |
		 DnDConstants.ACTION_LINK);
	}

	/**
	Construct a DndAppDisplay.
	@param world Host world serving as the root for all scene
	picking and object instantiation.
	@param dragSourceList List containing references to nodes
	that can act as drag source objects.
	@param dropTargetList List containing references to nodes
	that can act as drop target objects.
	@param dropSourceList List containing references to nodes
	that can act as drop source objects.
	@param dragActions Flags specifying the types of drag actions
	this object will provide (DnDConstants.ACTION_???).
	@param dropActions Flags specifying the types of drop actions
	this object will accept (DnDConstants.ACTION_???).
	*/
	public DndAppDisplay(AppWorld world,
	 ArrayList dragSourceList, ArrayList dropTargetList,
	 ArrayList dropSourceList, int dragActions, int dropActions){

		_world = world;

		// setup dragging
		if(dragSourceList != null) {				
			_dragSourceList = dragSourceList;
			
			_dragSource = new DragSource();

			_dragGesture =
			 _dragSource.createDefaultDragGestureRecognizer(
			 this, dragActions, this);
			
			_dragSourcePicker = new PickEngine(
			 _world.getSceneRoot(), _dragSourceList);
		}
		
		// setup dropping
		if(dropTargetList != null && dropSourceList != null) {				
			_dropSourceList = dropSourceList;
			
			_dropTarget = new DropTarget(
			 this, dropActions, this);
			 
			_dropTargetPicker = new PickEngine(
			 _world.getSceneRoot(), dropTargetList);
		}
	}

	// DragGestureListener implementation
	 
	public void dragGestureRecognized(DragGestureEvent evt) {
	
if(Debug.getEnabled()){
Debug.println(this, "DndAppDisplay.gesture",
"GESTURE:dragGestureRecognized:" +
" event=" + DndUtils.toString(evt));}

		// if over valid source, start drag
		DndNodeInfo source = pickDragSource(
		 evt.getDragOrigin().x, evt.getDragOrigin().y);
		
if(Debug.getEnabled()){
Debug.println(this, "DndAppDisplay.gesture",
"GESTURE:dragGestureRecognized:" +
" source=" + source);}

		if(source != null) {
			// get drag source object name
			StringSelection text = new StringSelection(
			 source.getName());
			
			// use name as transfer data
			_dragSource.startDrag(evt,
			 DragSource.DefaultCopyNoDrop, text, this);
		}
	}

	// DragSourceListener implementation
	 
	public void dragDropEnd(DragSourceDropEvent evt) {}

	public void dragEnter(DragSourceDragEvent evt) {}

	public void dragExit(DragSourceEvent evt) {}

	public void dragOver(DragSourceDragEvent evt) {}

	public void dropActionChanged(DragSourceDragEvent evt) {}

	// DropTargetListener implementation
	 
	public void dragEnter(DropTargetDragEvent evt) {
	
if(Debug.getEnabled()){
Debug.println(this, "DndAppDisplay.target",
"TARGET:dragEnter:" +
" event=" + DndUtils.toString(evt));}

		dragOver(evt);
	}

	public void dragExit(DropTargetEvent evt) {}

	public void dragOver(DropTargetDragEvent evt) {
	
if(Debug.getEnabled()){
Debug.println(this, "DndAppDisplay.target",
"TARGET:dragOver:" +
" event=" + DndUtils.toString(evt));}

		// if over valid target, accept drag
		DndNodeInfo target = pickDropTarget(
		 evt.getLocation().x, evt.getLocation().y);
		
		if(target != null) {
			evt.acceptDrag(evt.getDropAction());
		} else {
			evt.rejectDrag();
		}
	}

	public void drop(DropTargetDropEvent evt) {
	
if(Debug.getEnabled()){
Debug.println(this, "DndAppDisplay.target.drop",
"TARGET:drop:" +
" event=" + DndUtils.toString(evt));}

		// if unsupported action, no drop
		if(evt.getDropAction()==DnDConstants.ACTION_NONE) {
		
if(Debug.getEnabled()){
Debug.println(this, "DndAppDisplay.target.drop",
"TARGET:drop:" +
" No action.");}

			evt.rejectDrop();
			return;
		}
		
		// if unsupported transfer data, no drop
		Transferable tr = evt.getTransferable();
		if(!tr.isDataFlavorSupported(DataFlavor.stringFlavor)) {
		
if(Debug.getEnabled()){
Debug.println(this, "DndAppDisplay.target.drop",
"TARGET:drop:" +
" Bad flavor.");}

			evt.rejectDrop();
			return;
		}

		// good drop, do it		
		try {
			evt.acceptDrop(evt.getDropAction());

			// user transfer data as drop source object name
			String name = (String)tr.getTransferData(
			 DataFlavor.stringFlavor);
			
			// add source object at hit point in world			
			Node source = buildDropSource(evt.getLocation().x,
			 evt.getLocation().y, name);

			// add source object to list so draggable
			if(source != null) _dragSourceList.add(source);

			// done with drop			
			evt.getDropTargetContext().dropComplete(true);
		} catch (IOException io) {
			io.printStackTrace();
			evt.rejectDrop();
		} catch (UnsupportedFlavorException ufe) {
			ufe.printStackTrace();
			evt.rejectDrop();
		}
	}

	public void dropActionChanged (DropTargetDragEvent evt) {}
			
	// personal body ============================================

	private AppWorld _world;		
	private PickEngine _dragSourcePicker = null;		
	private PickEngine _dropTargetPicker = null;		
	private ArrayList _dragSourceList = null;		
	private ArrayList _dropSourceList = null;		
	private Vector2d _dispPos = new Vector2d();		
	private Point3d _hitPoint = new Point3d();		
	private DragGestureRecognizer _dragGesture;
	private DragSource _dragSource;
	private DropTarget _dropTarget;

	/**
	Performs a pick operation in the scene at the specified
	raw display position.  If a drag source object is hit its
	DnD info object is returned.
	@param posX Raw display X position, as from AWT.
	@param posY Raw display Y position, as from AWT.
	@return DnD info for the picked drag source object.  Null if
	none hit.
	*/
	protected DndNodeInfo pickDragSource(int posX, int posY) {
		
		// find over target, null if none
		int targetI = _dragSourcePicker.pickTarget(
		 this, Input.cookMouse(posX, posY, _dispPos), null);
		Node target = (Node)_dragSourcePicker.getPickTarget(
		 targetI);
		
if(Debug.getEnabled()){
Debug.println(this, "DndAppDisplay.over",
"DndAppDisplay.pickDragSource:" +
" target=" + target);}
		 
		if(target == null) return null;

		// if drag source, return it
		Object info = target.getUserData();
		
if(Debug.getEnabled()){
Debug.println(this, "DndAppDisplay.over",
"DndAppDisplay.pickDragSource" +
" info=" + info);}
		 
		if(info instanceof DndNodeInfo)
			return (DndNodeInfo)info;
		else
			return null;
	}

	/**
	Performs a pick operation in the scene at the specified
	raw display position.  If a drop target object is hit its
	DnD info object is returned.
	@param posX Raw display X position, as from AWT.
	@param posY Raw display Y position, as from AWT.
	@return DnD info for the picked drop target object.  Null if
	none hit.
	*/
	protected DndNodeInfo pickDropTarget(int posX, int posY) {	
		// find over target, null if none
		int targetI = _dropTargetPicker.pickTarget(
		 this, Input.cookMouse(posX, posY, _dispPos), null);
		Node target = (Node)_dropTargetPicker.getPickTarget(
		 targetI);
		
if(Debug.getEnabled()){
Debug.println(this, "DndAppDisplay.over",
"DndAppDisplay.pickDropTarget" +
" target=" + target);}
		 
		if(target == null) return null;

		// if drop target, return it
		Object info = target.getUserData();
		
if(Debug.getEnabled()){
Debug.println(this, "DndAppDisplay.over",
"DndAppDisplay.pickDropTarget" +
" info=" + info);}
		 
		if(info instanceof DndNodeInfo)
			return (DndNodeInfo)info;
		else
			return null;
	}
	
	/**
	Performs a pick operation in the scene at the specified
	raw display position.  If a drop target object is hit, a
	drop source object with the specified name will be created
	and added to the world scene at the hit point.
	@param posX Raw display X position, as from AWT.
	@param posY Raw display Y position, as from AWT.
	@return Reference to the object that was added to the scene.
	Null if none.
	*/
	protected Node buildDropSource(int posX, int posY,
	 String name) {

		// find named object in list, if none quit
		Node source = null;
		
		Iterator iter = _dropSourceList.iterator();
		while(iter.hasNext()) {
			Node node = (Node)iter.next();
			Object info = node.getUserData();
			if(info instanceof DndNodeInfo &&
			 name.equals(((DndNodeInfo)info).getName())) {
			 	source = node;
			 	break;
			} 
		}
		
if(Debug.getEnabled()){
Debug.println(this, "DndAppDisplay.over",
"DndAppDisplay.buildDropSource" +
" source=" + source);}
		 
		if(source == null) return null;
			
		// find hit point, if none quit
		int targetI = _dropTargetPicker.pickTarget(
		 this, Input.cookMouse(posX, posY, _dispPos), _hitPoint);
		
if(Debug.getEnabled()){
Debug.println(this, "DndAppDisplay.over",
"DndAppDisplay.buildDropSource" +
" targetI=" + targetI + " hitPoint=" + _hitPoint);}
		 
		if(targetI < 0) return null;

		// instantiate object in world
		_world.setLive(false);
		
		/// build transform at hit point
		Transform3D xform = new Transform3D();
		xform.set(new Vector3d(_hitPoint));

		/// build transform group with source object		
		TransformGroup group = new TransformGroup(xform);
		source = source.cloneTree();
		group.addChild(source);
		
		/// add object to world
		_world.addSceneNode(group);
		_world.setLive(true);
		
		return source;
	}

}