import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import j3dui.utils.Debug;
import j3dui.utils.*;
import j3dui.utils.app.*;
import j3dui.utils.objects.*;
import j3dui.utils.blocks.Blocks;

/**
Fancy application frame containing an AppWorld, two AppViews and
their displays, a data tree, and a data palette.  It has a menu
bar that includes functions for clearing the scene, loading new
models, and printing the displays.  The 2D UI components are all
done mostly using Swing components.
<P>
To use this frame...
<UL>
<LI> Get the world with getWorld() and add some visible objects
to its scene root.
<LI> Get the views with getViewLeft() and getViewRight() and
add them to the world's view root.
<LI> Use showFrame() when everything is ready to be shown.
</UL>
<P><B>Notes:</B><BR>
There appears to be a bug in Canvas3D that prevents capturing
the image of the second instance of a Canvas3D.  As such, the
left view always appears blank when trying to print it.  

@author Jon Barrilleaux,
copyright (c) 1999 Jon Barrilleaux,
All Rights Reserved.
*/
public class FancyFrame extends JFrame {
	
	// public interface =========================================
	
	/**
	Constructs a FancyFrame with an empty virtual world.
	*/
	public FancyFrame() {
		// kill the window on close
		addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent winEvent) {
				System.exit(0);
			}
		});
		
		// build dialogs
		String catDir = Blocks.buildResourcePath("catalog");
		
		_selectModelDialog = new FileDialog(this,
		 "Select Model", FileDialog.LOAD);
		_selectModelDialog.setDirectory(catDir);

		// tell Swing to use heavyweight popups	 
		JPopupMenu.setDefaultLightWeightPopupEnabled(false);
		ToolTipManager.sharedInstance().
		 setLightWeightPopupEnabled(false);
		 		
		// build Swing window components
		buildMenuBar();
		buildContentPanel();
	}

	/**
	Makes the world live and sizes and shows the frame.
	*/
	public void showFrame() {
		_tree.updateTree(_world.getSceneRoot());
		_world.setLive(true);
		
		// size contents THEN show frame
		pack();
		show();
	}

	/**
	Gets the frame's 3D world.
	@return The 3D world.
	*/
	public AppWorld getWorld() {
		return _world;
	}

	/**
	Gets the frame's left view.
	@return The left view.
	*/
	public AppView getViewLeft() {
		return _viewL;
	}

	/**
	Gets the frame's right view.
	@return The right view.
	*/
	public AppView getViewRight() {
		return _viewR;
	}

	/**
	Gets the frame's tree pane.
	@return The tree pane.
	*/
	public FancyTree getTree() {
		return _tree;
	}

	/**
	Gets the frame's palette pane.
	@return The palette pane.
	*/
	public FancyPalette getPalette() {
		return _palette;
	}
			
	// personal body ============================================

	/** The virtual world. */
	private AppWorld _world = new AppWorld();

	/** The window left view. */
	private AppView _viewL = new AppView();

	/** The window right view. */
	private AppView _viewR = new AppView();

	/** The window tree. */
	private FancyTree _tree;

	/** The window palette. */
	private FancyPalette _palette;
	
	/** Dialog for selecting model file to be loaded. */
	private FileDialog _selectModelDialog;
	
	/** Default value.  If true then newly created TestThings
	will be rendered using its "fast" shape. */
	private boolean _fastThings = false;
			
	/**
	Loads a model file and updates the frame.
	*/
	protected void loadModel(String fileName) {
		Node node = ModelLoader.loadModelFromName(fileName);
		if(node == null) return;

		// kill the scene before adding node or updating tree
		boolean isLive = _world.setLive(false);				
		_world.addSceneNode(node);
		_tree.updateTree(_world.getSceneRoot());
		_world.setLive(true);
	}

	// component builders
			
	/**
	Builds the window menubar and adds it to the window.
	*/
	protected void buildMenuBar() {
		
		// build the menu bar, by menu
		JMenuBar menuBar = new JMenuBar();
		
		JMenu menu;
		JMenuItem menuItem;
		JCheckBoxMenuItem checkItem;
		
		/// build the File menu, by item
		menu = new JMenu("File");
		
		menuItem = new JMenuItem("Clear Scene");
		menuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				boolean isLive = _world.setLive(false);				
				_world.clearScene();
				_tree.updateTree(_world.getSceneRoot());
				_palette.updatePalette();
				_world.setLive(true);				
			}
		});
		menu.add(menuItem);
		
		menuItem = new JMenuItem("Add Model");
		menuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				_selectModelDialog.show();
				if(_selectModelDialog.getFile() == null) return;
				
				String fileName = new String(
				 _selectModelDialog.getDirectory() +
				 _selectModelDialog.getFile());

				loadModel(fileName);
			}
		});
		menu.add(menuItem);
		
		menu.insertSeparator(menu.getItemCount());
		
		menuItem = new JMenuItem("Print Left");
		menuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				new PrintFrame(new ImageCanvas(
				 _viewL.getDisplay()));
			}
		});
		menu.add(menuItem);
		
		menuItem = new JMenuItem("Print Right");
		menuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				new PrintFrame(new ImageCanvas(
				 _viewR.getDisplay()));
			}
		});
		menu.add(menuItem);
		
		menu.insertSeparator(menu.getItemCount());
		
		menuItem = new JMenuItem("Exit");
		menuItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				System.exit(0);
			}
		});
		menu.add(menuItem);
		
		menuBar.add(menu);
		
		/// build the Options menu, by item

		menu = new JMenu("Options");
		
		TestThing.setFast(_fastThings); // init target
		checkItem = new JCheckBoxMenuItem(
		 "Fast Things", _fastThings);
		checkItem.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent event) {
				_fastThings =
				 event.getStateChange()==ItemEvent.SELECTED;
				TestThing.setFast(_fastThings);
			}
		});
		menu.add(checkItem);
		
		menuBar.add(menu);

		// add menu bar to frame		
		getRootPane().setJMenuBar(menuBar);
	}
			
	/**
	Builds the window contents panel, which consists of a
	view panel, a data tree panel, and a data palette panel,
	and adds it to the window.
	*/
	protected void buildContentPanel() {
		// build main panel
		JSplitPane panel = new JSplitPane(
		 JSplitPane.VERTICAL_SPLIT);
		panel.setDoubleBuffered(true);
		panel.setContinuousLayout(true);
		panel.setOneTouchExpandable(true);

		// build view subpanel
		JPanel viewPanel = new JPanel();
		panel.setTopComponent(viewPanel);
		
		viewPanel.setPreferredSize(new Dimension(400, 200));
		viewPanel.setLayout(new GridLayout(1, 2, 2, 2));
		
		viewPanel.add(_viewL.getDisplay());
		viewPanel.add(_viewR.getDisplay());
		
		// build data subpanel
		JSplitPane dataPanel = new JSplitPane(
		 JSplitPane.HORIZONTAL_SPLIT);
		panel.setBottomComponent(dataPanel);
		
		dataPanel.setPreferredSize(new Dimension(400, 150));
		dataPanel.setDoubleBuffered(true);
		dataPanel.setContinuousLayout(true);
		dataPanel.setOneTouchExpandable(true);
		
		/// build data tree
		_tree = new FancyTree();
		dataPanel.setLeftComponent(_tree);
		
		/// build data palette
		_palette = new FancyPalette(this);
		dataPanel.setRightComponent(_palette);

		// add main panel to frame		
		getContentPane().add(panel, BorderLayout.CENTER);
	}
	
}