import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import j3dui.utils.Debug;
import j3dui.utils.*;
import j3dui.utils.app.*;
import j3dui.utils.objects.*;
import j3dui.utils.blocks.*;
import j3dui.control.*;
import j3dui.control.inputs.*;
import j3dui.control.inputs.sensors.*;
import j3dui.control.mappers.*;
import j3dui.control.mappers.intuitive.*;
import j3dui.control.actuators.*;
import j3dui.control.actuators.groups.*;
import j3dui.feedback.*;
import j3dui.feedback.elements.*;
import j3dui.visualize.*;
import j3dui.visualize.change.*;

/**
Example demonstrating third person manipulation
with perfectly overlaid target decorations.
<P>
Refer to the book for further details about this example.

@author Jon Barrilleax,
copyright (c) 1999-2000 Jon Barrilleaux,
All Rights Reserved.
*/

public class ThirdPerson extends SingleFrame {
	
	public static void main(String[] argv) {
		Debug.loadAllProperties();
		new ThirdPerson();
	}
	
	// public interface =========================================
	
	public ThirdPerson() {
		setTitle("Example - ThirdPerson");
		
		new HelpDialog(this, "Help - ThirdPerson",
		 "Example demonstrating third person manipulation " +
		 "with perfectly overlaid target decorations." +
		 "\n" +
		 "\n" +
		 "> Resize display window." +
		 "\n" +
		 "> Left-button-shift drag or arrow keys-shift to " +
		 "tilt and turn view." +
		 "\n" +
		 "> Single/double/triple click things to enable " +
		 "sliding/rotating/lifting them." +
		 "\n" +
		 "> Single/double/triple-ctrl click things to " +
		 "group enable them." +
		 "\n" +
		 "> Left-button drag or use arrow keys to manipulate " +
		 "selected thing or group." +
		 "\n" +
		 "> Escape key to cancel selection and manipulation." +
		 "\n" +
		 "> Turn rotate control to turn things selected for " +
		 "rotation." +
		 "\n" +
		 "> Slide lift control to lift things selected for " +
		 "lifting." +
		 "\n" +
		 "\n" +
		 ">>> Mouse-over, select, and/or drag for visual " +
		 "and audio feedback." +
		 "\n" +
		 ">>> Display decorations remain perfectly " +
		 "overlaid  regardless of how things, view, or " +
		 "display is manipulated." +
		 "\n" +
		 "\n" +
		 "Refer to the book for further details about this " +
		 "example." +
		 "\n" +
		 "copyright (c) 1999-2000 Jon Barrilleaux, " +
		 "All Rights Reserved."
		);

		// setup orbitable view, first+shift
		getWorld().addViewNode(
		 ActuationBlocks.buildViewOrbiter(getView(),
		 new Vector3d(-Math.PI/6, -Math.PI/8, 0), 10,
		 Input.BUTTON_FIRST, Input.MODIFIER_SHIFT,
		 Input.MODIFIER_SHIFT)
		);

		// build view change sensor
		ViewChangeSplitter viewSpltr = new ViewChangeSplitter();
		
		new ViewExternalChangeSensor(viewSpltr, getView());
		new ViewInternalChangeSensor(viewSpltr, getView());
		
		// build target draggers
		/// absolute draggers (for WRM slide)
		InputDragSplitter absDrag = new InputDragSplitter();
		
		ActuationBlocks.buildAbsoluteDragger(absDrag,
		 getView(), Input.BUTTON_FIRST,
		 Input.MODIFIER_NONE, Input.MODIFIER_NONE);
		
		/// relative draggers (for direct rotate & lift)
		InputDragSplitter relDrag = new InputDragSplitter();
		
		ActuationBlocks.buildRelativeDragger(relDrag,
		 getView(), Input.BUTTON_FIRST,
		 Input.MODIFIER_NONE, Input.MODIFIER_NONE);
		
		// build feedback manager
		FeedbackGroupManager groupMgr =
		 FeedbackBlocks.buildGroupManager(
		 getView().getDisplay(), getWorld().getViewRoot());

		// build room with floor
		SecondBlocks.Room room = new SecondBlocks.Room(
		 new Vector3d(8, 3, 8));
		getWorld().addSceneNode(room);

		/// use room for deselect
		FeedbackBlocks.buildTargetFeedbackSupport(room,
		 groupMgr, relDrag, getView().getDisplay(),
		 getWorld().getSceneRoot());
				
		// build targets
		Node thing;
		
		/// lamp
		thing = ModelLoader.loadModelFromName(
		 Blocks.buildResourcePath("catalog/lamp.wrl"));
		
		SecondBlocks.FloorTarget lamp = new SecondBlocks.
		 FloorTarget(thing, room, viewSpltr, absDrag, relDrag,
		 groupMgr, getWorld(), getView());
		lamp.getSlide().initActuation(
		 new Vector4d(-.45, 0, -.3, 0));
		lamp.getLift().initActuation(
		 new Vector4d(0, 1.65, 0, 0));
		getWorld().addSceneNode(lamp);

		/// desk
		thing = ModelLoader.loadModelFromName(
		 Blocks.buildResourcePath("catalog/desk.wrl"));
		
		SecondBlocks.FloorTarget desk = new SecondBlocks.
		 FloorTarget(thing, room, viewSpltr, absDrag, relDrag,
		 groupMgr, getWorld(), getView());
		getWorld().addSceneNode(desk);

		/// drawers
		thing = ModelLoader.loadModelFromName(
		 Blocks.buildResourcePath("catalog/drawers.wrl"));
		
		SecondBlocks.FloorTarget drawers = new SecondBlocks.
		 FloorTarget(thing, room, viewSpltr, absDrag, relDrag,
		 groupMgr, getWorld(), getView());
		drawers.getSlide().initActuation(
		 new Vector4d(1.2, 0, .05, 0));
		drawers.getLift().initActuation(
		 new Vector4d(0, .55, 0, 0));
		getWorld().addSceneNode(drawers);
		
		// build control panel
		Vector2d size = new Vector2d();
		Vector2d limits = new Vector2d();
		Vector2d origin = new Vector2d();
		Vector2d offset = new Vector2d();
		double scale;
		Node mark, back, frame;
		
		/// rotate control
		size.set(66, 66);
		limits.set(0, -1); // ignore limits
		scale = 1.0;
		origin.set(-.45, -.45);
		offset.set(size.x/2.0, size.y/2.0);
		
		mark = ThirdBlocks.buildMultiMark(true, 10, 10);
		back = ThirdBlocks.buildMultiRotateBack(true,
		 size.x, size.y);
		frame = new TextureRect(false, new Color3f(.5f,.5f,.5f),
		 .5, null, Elements.SIDE_TOP, size.x, size.y);
		
		ThirdBlocks.RotateControl rotate = new ThirdBlocks.
		 RotateControl(mark, back, frame, origin, offset,
		 new Vector2d(-21, 0), new Vector2d(0, 6), limits,
		 scale, viewSpltr, getView());
		
		/// lift control
		size.set(66, 66);
		limits.set(0, 38);
		scale = room.getSize().y / (limits.y - limits.x);
		origin.set(.45, -.45);
		offset.set(-size.x/2.0, size.y/2.0);
		
		mark = ThirdBlocks.buildMultiMark(true, 10, 10);
		back = ThirdBlocks.buildMultiLiftBack(true,
		 size.x, size.y);
		frame = new TextureRect(false, new Color3f(.5f,.5f,.5f),
		 .5, null, Elements.SIDE_TOP, size.x, size.y);
		
		ThirdBlocks.LiftControl lift = new ThirdBlocks.
		 LiftControl(mark, back, frame, origin, offset,
		 new Vector2d(0, 0), new Vector2d(0, -12), limits,
		 scale, viewSpltr, getView());
		
		// connect controls and targets
		ActuationSplitter enabler;
		FeedbackTrigger filter;
		FeedbackEnableTrigger trigger;
		
		/// interconnect control-target actuation
		ThirdBlocks.connectRotateControl(rotate, lamp);
		ThirdBlocks.connectRotateControl(rotate, desk);
		ThirdBlocks.connectRotateControl(rotate, drawers);
		
		ThirdBlocks.connectLiftControl(lift, lamp);
		ThirdBlocks.connectLiftControl(lift, desk);
		ThirdBlocks.connectLiftControl(lift, drawers);
		
		/// enable rotate control if target selected for rotate
		trigger = new FeedbackEnableTrigger(rotate);
		trigger.setSelectFlags(Feedback.SELECT_DOUBLE);
		
		lamp.addEventTarget(trigger);
		desk.addEventTarget(trigger);
		drawers.addEventTarget(trigger);
		
		/// enable lift control if target selected for lift
		trigger = new FeedbackEnableTrigger(lift);
		trigger.setSelectFlags(Feedback.SELECT_TRIPLE);
		
		lamp.addEventTarget(trigger);
		desk.addEventTarget(trigger);
		drawers.addEventTarget(trigger);
		
		// initialize feedback (last)
		groupMgr.initTargets(null, Feedback.STATUS_DISABLE,
		 Feedback.SELECT_NORMAL, Feedback.ACTION_NORMAL);

		// done
		showFrame();
	}
			
	// personal body ============================================

}